var Driver = /** @class */ (function () {
    function Driver() {
        if (this.constructor === Driver) {
            throw new TypeError('Illegal constructor');
        }
    }
    Driver.prototype.cancel = function (promise) {
        throw new Error('Method "Cancel" is unimplemented!!');
    };
    Driver.prototype.register = function (name) {
        DriverManager.register(name, this);
    };
    return Driver;
}());

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = function(d, b) {
    extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return extendStatics(d, b);
};

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
}

function __awaiter(thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

var LazyDriver = /** @class */ (function (_super) {
    __extends(LazyDriver, _super);
    function LazyDriver(name) {
        var _this = _super.call(this) || this;
        _this._promises = [];
        _this._interval = setInterval(_this.checkOnLoad.bind(_this), 500);
        _this._showLoadWarningOnlyOnce = false;
        _this._loadWarningTime = 3000;
        _this._loadErrorTime = 10000;
        _this._benchmarking = performance.now();
        _this._name = name;
        return _this;
    }
    Object.defineProperty(LazyDriver.prototype, "benchmarking", {
        get: function () {
            return performance.now() - this._benchmarking;
        },
        enumerable: true,
        configurable: true
    });
    LazyDriver.prototype.load = function (options) {
        return __awaiter(this, void 0, void 0, function () {
            var _this = this;
            return __generator(this, function (_a) {
                if (!this._promises[this._name]) {
                    this._promises[this._name] = [];
                }
                return [2 /*return*/, new Promise(function (resolve, reject) {
                        _this._promises[_this._name].push(function (driver) {
                            driver.load(options).then(resolve, reject);
                        });
                    })];
            });
        });
    };
    LazyDriver.prototype.process = function (driver) {
        if (driver instanceof LazyDriver) {
            return false;
        }
        if (this._promises[this._name]) {
            this._promises[this._name].forEach(function (callback) { return callback(driver); });
            delete this._promises[this._name];
        }
        clearInterval(this._interval);
    };
    LazyDriver.prototype.checkOnLoad = function () {
        var bench = this.benchmarking;
        if (bench >= this._loadWarningTime &&
            bench < this._loadErrorTime &&
            !this._showLoadWarningOnlyOnce) {
            this._showLoadWarningOnlyOnce = true;
            console.warn("Driver \"" + this._name + "\" took more than " + this._loadWarningTime / 1000 + " seg to load");
        }
        else if (bench > this._loadWarningTime && bench > this._loadErrorTime) {
            clearInterval(this._interval);
            console.error("Driver \"" + this._name + "\" took more than " + this._loadErrorTime / 1000 + " seg to load");
        }
    };
    return LazyDriver;
}(Driver));

var UrlDriver = /** @class */ (function (_super) {
    __extends(UrlDriver, _super);
    function UrlDriver() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    UrlDriver.prototype.cancel = function (promise) {
        if (promise && promise.abort) {
            promise.abort();
            return true;
        }
        return false;
    };
    UrlDriver.prototype.load = function (options) {
        var params = {
            method: options.method,
            data: Object.assign({}, options.params, options.extraParams),
        };
        return Http.request(options.url, params);
    };
    return UrlDriver;
}(Driver));

var FilterOperator;
(function (FilterOperator) {
    FilterOperator["Equal"] = "=";
    FilterOperator["Equality"] = "==";
    FilterOperator["NotEqual"] = "!=";
    FilterOperator["GreaterThan"] = ">";
    FilterOperator["GreaterThanEqualTo"] = ">=";
    FilterOperator["LessThan"] = "<";
    FilterOperator["LessThanEqualTo"] = "<=";
    FilterOperator["Contains"] = "%%";
})(FilterOperator || (FilterOperator = {}));
function filterCb(array, filters) {
    return filters && array && array.filter(function (item) {
        var remove = false;
        filters.forEach(function (filter) {
            var value = filter.value;
            if (value == undefined ||
                value == '') {
                return false;
            }
            var value2String = value.toString().toLowerCase();
            var keys = filter.field.split("|");
            var fields = keys.map(function (key) { return item[key] && item[key].toString().toLowerCase() || ""; });
            switch (filter.operator) {
                case FilterOperator.Equal:
                case FilterOperator.Equality:
                    if (!(fields.find(function (item) { return item == value2String; })))
                        remove = true;
                    break;
                case FilterOperator.NotEqual:
                    if (!(fields.find(function (item) { return item != value2String; })))
                        remove = true;
                    break;
                case FilterOperator.GreaterThan:
                    if (!(fields.find(function (item) { return item > value2String; })))
                        remove = true;
                    break;
                case FilterOperator.GreaterThanEqualTo:
                    if (!(fields.find(function (item) { return item >= value2String; })))
                        remove = true;
                    break;
                case FilterOperator.LessThan:
                    if (!(fields.find(function (item) { return item < value2String; })))
                        remove = true;
                    break;
                case FilterOperator.LessThanEqualTo:
                    if (!(fields.find(function (item) { return item <= value2String; })))
                        remove = true;
                    break;
                case FilterOperator.Contains:
                    if (!(fields.find(function (item) { return item.indexOf(value2String) > -1; })))
                        remove = true;
                    break;
                default:
                    if (!(fields.find(function (item) { return item == value2String; })))
                        remove = true;
            }
            if (remove)
                return false;
        });
        return !remove;
    });
}

function groupCb(array, groupField) {
    if (groupField) {
        var groups_1 = new Map();
        array.forEach(function (item) {
            var group = item[groupField];
            if (!groups_1.has(group)) {
                groups_1.set(group, []);
            }
            groups_1.get(group).push(item);
        });
        return Array.from(groups_1.values());
    }
    else {
        return array;
    }
}

function pruneCb(array, prune) {
    return array && prune && array.slice(prune.start, prune.limit + (prune.start || 0));
}

function sorterCb(array, sorter) {
    var direction = sorter.direction;
    return array && sorter && array.sort(function (a, b) {
        var aField = a[sorter.field];
        var bField = b[sorter.field];
        if (aField === bField)
            return 0;
        return direction === "DESC" ? (aField < bField ? 1 : -1) : (aField > bField ? 1 : -1);
    });
}

var Http = /** @class */ (function () {
    function Http() {
    }
    Http.request = function (url, params) {
        var promise, xhr, fnResolve, fnReject, fnAbort;
        var method = params.method || 'GET';
        var async = params.async != false;
        var user = params.user || '';
        var password = params.password || '';
        var qs = '';
        var body;
        if (['GET', 'DELETE'].indexOf(method) > -1)
            qs = (params.data && url.includes("?") ? '&' : '?') + Http.queryString(params.data || {});
        else // POST or PUT
            body = JSON.stringify(params.data);
        url = url + qs;
        promise = new Promise(function (resolve, reject) {
            xhr = new XMLHttpRequest();
            xhr.open(method, url, async, user, password);
            xhr.onload = function () {
                if (this.status >= 200 && this.status < 300) {
                    resolve(xhr.response);
                }
                else {
                    reject({
                        status: this.status,
                        statusText: xhr.statusText
                    });
                }
            };
            xhr.responseType = params.responseType || 'json';
            if (params && params.headers) {
                Object.keys(params.headers).forEach(function (key) {
                    if (params.headers) {
                        xhr.setRequestHeader(key, params.headers[key]);
                    }
                });
            }
            xhr.send(body);
            fnResolve = resolve;
            fnReject = reject;
            fnAbort = xhr.abort;
        });
        promise.resolve = fnResolve;
        promise.reject = fnReject;
        promise.abort = fnAbort.bind(xhr);
        return promise;
    };
    Http.get = function (url, params) { return Http.request(url, Object.assign({ method: 'GET' }, params)); };
    Http.post = function (url, params) { return Http.request(url, Object.assign({ method: 'POST' }, params)); };
    Http.put = function (url, params) { return Http.request(url, Object.assign({ method: 'PUT' }, params)); };
    Http.delete = function (url, params) { return Http.request(url, Object.assign({ method: 'DELETE' }, params)); };
    Http.queryString = function (params) {
        var euc = encodeURIComponent;
        return Object
            .keys(params)
            .map(function (k) {
            if (Array.isArray(params[k])) {
                return params[k]
                    .map(function (val) { return euc(k) + "[]=" + euc(val); })
                    .join('&');
            }
            return euc(k) + "=" + euc(params[k]);
        })
            .join('&');
    };
    return Http;
}());

var EventManager = /** @class */ (function () {
    function EventManager() {
        this.listeners = [];
        this.listenersOncer = [];
    }
    EventManager.prototype.on = function (event, cb) {
        if (!this.listeners[event]) {
            this.listeners[event] = [];
        }
        this.listeners[event].push(cb);
    };
    EventManager.prototype.off = function (event, cb) {
        if (this.listeners[event]) {
            this.listeners[event] = this.listeners[event].filter(function (callback) { return callback !== cb; });
            if (this.listeners[event].length == 0) {
                delete this.listeners[event];
            }
        }
    };
    EventManager.prototype.once = function (event, cb) {
        if (!this.listenersOncer[event]) {
            this.listenersOncer[event] = [];
        }
        this.listenersOncer[event].push(cb);
    };
    EventManager.prototype.fire = function (event, parameters) {
        var eventDetail = Object.assign({
            event: event,
            parent: this,
            timestamp: new Date().getTime(),
            parameters: parameters
        });
        /** Update general listeners */
        if (this.listeners[event]) {
            this.listeners[event].forEach(function (listener) {
                try {
                    listener(parameters, eventDetail);
                }
                catch (e) {
                    console.error(e);
                }
            });
        }
        /** Update Multi listeners */
        if (this.listeners["*"]) {
            this.listeners["*"].forEach(function (listener) {
                try {
                    listener(parameters, eventDetail);
                }
                catch (e) {
                    console.error(e);
                }
            });
        }
        /** Clear the `once` queue */
        if (this.listenersOncer[event]) {
            var eventListeners = this.listenersOncer[event];
            this.listenersOncer[event] = [];
            eventListeners.forEach(function (listener) {
                try {
                    listener(parameters, eventDetail);
                }
                catch (e) {
                    console.error(e);
                }
            });
        }
    };
    return EventManager;
}());

function DriverMap() {
    // create map
    let map = new Map();
    // set new prototype
    Object.setPrototypeOf(map, DriverMap.prototype);
    // return map with prototype driverMap
    return map;
}

DriverMap.__proto__ = Map;
DriverMap.prototype.__proto__ = Map.prototype;

Object.defineProperties(DriverMap.prototype, {
    get: {
        enumerable: false,
        value: function (key) {
            let instance = Map.prototype.get.call(this, key);
            if (instance) {
                instance = Object.create(instance);
            }
            return instance;
        }
    },
    set: {
        enumerable: false,
        value: function (key, value) {
            if (!(value instanceof Driver)) {
                throw new Error(`"${value.constructor.name}" isn't instance of Driver`)
            }
            if (this.observer[key]) {
                this.observer[key](value);
            }
            return Map.prototype.set.call(this, key, value);
        }
    },
    watch: {
        enumerable: false,
        value: function (key, callback) {
            this.observer[key] = callback;
        }
    },
    unwatch: {
        enumerable: false,
        value: function (key) {
            this.observer[key] = undefined;
            delete this.observer[key];
        }
    },
    observer: {
        enumerable: false,
        writable: true,
        value: [],
    }
});

var DriverManager = /** @class */ (function () {
    function DriverManager() {
        this.drivers = new DriverMap();
        this.drivers.set("url", new UrlDriver());
    }
    DriverManager.getInstance = function () {
        if (!this.instance) {
            this.instance = new DriverManager();
        }
        return this.instance;
    };
    DriverManager.destroy = function () {
        this.instance.drivers.clear();
        delete this.instance;
    };
    DriverManager.register = function (name, driver) {
        this.getInstance().drivers.set(name, driver);
    };
    DriverManager.registry = function (name) {
        var drivers = this.getInstance().drivers;
        if (drivers.has(name)) {
            return drivers.get(name);
        }
        else {
            // create lazy driver
            var lazydriver_1 = new LazyDriver(name);
            // register lazy driver
            lazydriver_1.register(name);
            // watch lazy driver
            drivers.watch(name, function (driver) {
                // process new driver
                lazydriver_1.process(driver);
                // unwatch lazy driver
                drivers.unwatch(name);
            });
            // return lazy driver
            return lazydriver_1;
        }
    };
    DriverManager.get = function (name) {
        var drivers = this.getInstance().drivers;
        return drivers.get(name);
    };
    return DriverManager;
}());

var ProxyPool = /** @class */ (function () {
    function ProxyPool() {
        this.pool = new Map();
    }
    ProxyPool.getInstance = function () {
        if (!this.instance) {
            this.instance = new ProxyPool();
        }
        return this.instance;
    };
    ProxyPool.destroy = function () {
        this.instance.pool = new Map();
        delete this.instance;
    };
    ProxyPool.prototype.getPool = function () {
        return this.pool;
    };
    ProxyPool.loadAll = function () {
        this.getInstance().getPool().forEach(function (proxy) {
            proxy.load();
        });
    };
    ProxyPool.load = function (code, options) {
        var pool = this.getInstance().getPool();
        var proxy = pool.get(code);
        if (proxy) {
            if (options) {
                proxy.options = options;
            }
            else {
                proxy.load();
            }
        }
    };
    ProxyPool.remove = function (code) {
        var pool = this.getInstance().getPool();
        return pool.delete(code);
    };
    ProxyPool.get = function (param) {
        var code;
        var pool = this.getInstance().getPool();
        if (typeof param == "object") {
            code = Proxy.getHash(param);
            if (!pool.has(code)) {
                param.id = code;
                pool.set(code, new Proxy(param));
            }
        }
        else {
            code = param;
        }
        return pool.get(code);
    };
    return ProxyPool;
}());

function isEqual(a, b) {
    if (a === b) return true;
    if (a && b && typeof a == 'object' && typeof b == 'object') {
        if (a.constructor !== b.constructor) return false;
        var length, i, keys;
        if (Array.isArray(a)) {
            length = a.length;
            if (length != b.length) return false;
            for (i = length; i-- !== 0;)
                if (!isEqual(a[i], b[i])) return false;
            return true;
        }
        if ((a instanceof Map) && (b instanceof Map)) {
            if (a.size !== b.size) return false;
            for (i of a.entries())
                if (!b.has(i[0])) return false;
            for (i of a.entries())
                if (!isEqual(i[1], b.get(i[0]))) return false;
            return true;
        }
        if ((a instanceof Set) && (b instanceof Set)) {
            if (a.size !== b.size) return false;
            for (i of a.entries())
                if (!b.has(i[0])) return false;
            return true;
        }
        if (ArrayBuffer.isView(a) && ArrayBuffer.isView(b)) {
            length = a.length;
            if (length != b.length) return false;
            for (i = length; i-- !== 0;)
                if (a[i] !== b[i]) return false;
            return true;
        }
        if (a.constructor === RegExp) return a.source === b.source && a.flags === b.flags;
        if (a.valueOf !== Object.prototype.valueOf) return a.valueOf() === b.valueOf();
        if (a.toString !== Object.prototype.toString) return a.toString() === b.toString();
        keys = Object.keys(a);
        length = keys.length;
        if (length !== Object.keys(b).length) return false;
        for (i = length; i-- !== 0;)
            if (!Object.prototype.hasOwnProperty.call(b, keys[i])) return false;
        for (i = length; i-- !== 0;) {
            var key = keys[i];
            if (!isEqual(a[key], b[key])) return false;
        }
        return true;
    }
    // true if both NaN, false otherwise
    return a!==a && b!==b;
}

var Proxy = /** @class */ (function (_super) {
    __extends(Proxy, _super);
    function Proxy(args) {
        var _this = _super.call(this) || this;
        _this._type = "url";
        _this._options = {};
        _this._data = [];
        _this._loading = false;
        _this._id = Proxy.getHash(args);
        _this._type = args.type || _this._type;
        _this._options = args;
        // registry driver type
        DriverManager.registry(_this.type);
        // load on init
        if (args.load)
            _this.load();
        return _this;
    }
    Object.defineProperty(Proxy.prototype, "id", {
        get: function () {
            return this._id;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Proxy.prototype, "type", {
        get: function () {
            return this._type;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Proxy.prototype, "error", {
        set: function (reason) {
            if (!isEqual(this._error, reason)) {
                this._error = reason;
                this.fire(Proxy.ERROR_EVENT, reason);
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Proxy.prototype, "loading", {
        set: function (loading) {
            if (!isEqual(this._loading, loading)) {
                this._loading = loading;
                this.fire(Proxy.LOADING_EVENT, loading);
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Proxy.prototype, "data", {
        get: function () {
            return this._data;
        },
        set: function (data) {
            var _this = this;
            if (data instanceof Object) {
                // force data to be array
                // doing for change JSON OBJECT(can Array[associated]) to Array,
                this._data = [];
                Array.from(Object.keys(data)).forEach(function (key) {
                    _this._data[key] = data[key];
                });
            }
            else {
                this._data = data;
            }
            this.fire(Proxy.LOAD_DATA_EVENT, data);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Proxy.prototype, "options", {
        get: function () {
            return this._options;
        },
        set: function (options) {
            if (!isEqual(this._options, options)) {
                this._options = options;
                this.load();
            }
        },
        enumerable: true,
        configurable: true
    });
    Proxy.getHash = function (config) {
        return config.id || Math.random().toString(36).substring(3);
    };
    Proxy.prototype.load = function () {
        return __awaiter(this, void 0, void 0, function () {
            var driver;
            var _this = this;
            return __generator(this, function (_a) {
                driver = DriverManager.get(this.type);
                if (this._promise) {
                    driver.cancel(this._promise);
                    delete this._promise;
                }
                this.loading = true;
                this._promise = driver.load(this.options);
                this._promise
                    .then(function (response) { return _this.data = response; })
                    .catch(function (reason) { return _this.error = reason; })
                    .finally(function () { _this.loading = false; delete _this._promise; });
                return [2 /*return*/];
            });
        });
    };
    Proxy.prototype.on = function (event, cb) {
        _super.prototype.on.call(this, event, cb);
        if (event == Proxy.ERROR_EVENT && this._error) {
            cb(this._error);
        }
        else if (event == Proxy.LOADING_EVENT && this._loading) {
            cb(this._loading);
        }
        else if (event == Proxy.LOAD_DATA_EVENT && this._data && this._data.length > 0) {
            cb(this._data);
        }
    };
    Proxy.LOAD_DATA_EVENT = 'loadData';
    Proxy.LOADING_EVENT = 'loading';
    Proxy.ERROR_EVENT = 'error';
    return Proxy;
}(EventManager));

var Source = /** @class */ (function () {
    function Source(args) {
        var _this = this;
        this._data = [];
        this._handleFilters = filterCb;
        this._handleSorter = sorterCb;
        this._handlePrune = pruneCb;
        this._handleGroup = groupCb;
        //handles Event's
        this._handleDataEvent = function (data) { return _this.data = data; };
        this._handleLoadingEvent = function (loading) { return _this.loading = loading; };
        this._handleErrorEvent = function (reason) { return _this.error = reason; };
        var filters = args.filters, sorter = args.sorter, prune = args.prune, group = args.group, paginate = args.paginate, handleDataChange = args.handleDataChange, handleError = args.handleError, handleLoading = args.handleLoading, children = args.children, options = __rest(args, ["filters", "sorter", "prune", "group", "paginate", "handleDataChange", "handleError", "handleLoading", "children"]);
        this._filters = filters;
        this._sorter = sorter;
        this._group = group;
        this._paginate = !!paginate;
        this._prune = paginate && !prune ? { start: 0, limit: 20 } : prune;
        this._handleDataChange = handleDataChange;
        this._handleError = handleError;
        this._handleLoading = handleLoading;
        var proxy = ProxyPool.get(options);
        this._proxyId = proxy.id;
        proxy.on(Proxy.LOAD_DATA_EVENT, this._handleDataEvent);
        proxy.on(Proxy.LOADING_EVENT, this._handleLoadingEvent);
        proxy.on(Proxy.ERROR_EVENT, this._handleErrorEvent);
    }
    Object.defineProperty(Source.prototype, "handleDataChange", {
        set: function (cb) {
            this._handleDataChange = cb;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "handleError", {
        set: function (cb) {
            this._handleError = cb;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "handleLoading", {
        set: function (cb) {
            this._handleLoading = cb;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "handleMapping", {
        set: function (cb) {
            this._handleMapping = cb;
            this.updateData();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "handleFilters", {
        set: function (cb) {
            this._handleFilters = cb;
            this.updateData();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "handleSorter", {
        set: function (cb) {
            this._handleSorter = cb;
            this.updateData();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "handlePrune", {
        set: function (cb) {
            this._handlePrune = cb;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "handleGroup", {
        set: function (cb) {
            this._handleGroup = cb;
        },
        enumerable: true,
        configurable: true
    });
    Source.prototype.applyMapping = function (data) {
        return this._handleMapping ? this._handleMapping(data) : data;
    };
    Source.prototype.applyFilters = function (data) {
        return this._filters && this._handleFilters ? this._handleFilters(data, this._filters) : data;
    };
    Source.prototype.applySorter = function (data) {
        return this._sorter && this._handleSorter ? this._handleSorter(data, this._sorter) : data;
    };
    Source.prototype.applyPrune = function (data) {
        return this._prune && this._handlePrune ? this._handlePrune(data, this._prune) : data;
    };
    Source.prototype.applyGroup = function (data) {
        return this._group && this._handleGroup ? this._handleGroup(data, this._group) : data;
    };
    Source.prototype.updateData = function () {
        var proxy = ProxyPool.get(this.proxyId);
        var dataMapping = this.applyMapping(proxy.data);
        var dataFiltered = this.applyFilters(dataMapping);
        var dataSortered = this.applySorter(dataFiltered);
        // let dataGrouped  = this.applyGroup(dataSortered);
        this._data = dataSortered;
    };
    Source.prototype.changeData = function () {
        this._handleDataChange &&
            this._handleDataChange(this.wrapper);
    };
    Source.prototype.resetPagination = function () {
        Object.assign(this._prune, { start: 0 });
    };
    Object.defineProperty(Source.prototype, "loading", {
        set: function (value) {
            this._handleLoading &&
                this._handleLoading(value);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "error", {
        set: function (value) {
            this._handleError &&
                this._handleError(value);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "data", {
        get: function () {
            return this.wrapper.data;
        },
        set: function (data) {
            this.updateData();
            this.changeData();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "wrapper", {
        get: function () {
            var wrapper = {
                data: this.applyPrune(this._data)
            };
            if (this._paginate === true && this._prune) {
                Object.assign(wrapper, this.pagination());
            }
            return wrapper;
        },
        enumerable: true,
        configurable: true
    });
    Source.prototype.pagination = function () {
        var _this = this;
        var length = this._data.length;
        var limit = this._prune && this._prune.limit || 20;
        var start = this._prune && this._prune.start || 0;
        return {
            total: length,
            currentPage: Math.floor(start / limit) + 1,
            lastPage: Math.floor(length / limit),
            perPage: limit,
            from: start,
            to: limit + start,
            previous: function () { Object.assign(_this._prune, { start: Math.max(start - limit, 0) }); _this.changeData(); },
            hasPreviousPage: function () { return start > 0; },
            next: function () { Object.assign(_this._prune, { start: Math.min(start + limit, Math.floor(length / limit) * limit) }); _this.changeData(); },
            hasNextPage: function () { return (limit + start) < length; },
        };
    };
    Object.defineProperty(Source.prototype, "filters", {
        set: function (value) {
            if (!isEqual(this._filters, value)) {
                this._filters = value;
                this.resetPagination();
                this.data = [];
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "sorter", {
        set: function (value) {
            if (!isEqual(this._sorter, value)) {
                this._sorter = value;
                this.resetPagination();
                this.data = [];
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "prune", {
        set: function (value) {
            if (!isEqual(this._prune, value)) {
                this._prune = value;
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "group", {
        set: function (value) {
            if (!isEqual(this._group, value)) {
                this._group = value;
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Source.prototype, "proxyId", {
        get: function () {
            return this._proxyId;
        },
        enumerable: true,
        configurable: true
    });
    Source.prototype.remove = function () {
        var proxy = ProxyPool.get(this.proxyId);
        proxy.off(Proxy.LOAD_DATA_EVENT, this._handleDataEvent);
        proxy.off(Proxy.LOADING_EVENT, this._handleLoadingEvent);
        proxy.off(Proxy.ERROR_EVENT, this._handleErrorEvent);
    };
    return Source;
}());



var Mordelon = /*#__PURE__*/Object.freeze({
    __proto__: null,
    ProxyPool: ProxyPool,
    EventManager: EventManager,
    DriverManager: DriverManager,
    Proxy: Proxy,
    Driver: Driver,
    Source: Source
});

export default Mordelon;
